//------------------------------------------------------------------------
//  EDIT LOOP
//------------------------------------------------------------------------
//
//  Eureka DOOM Editor
//
//  Copyright (C) 2001-2009 Andrew Apted
//  Copyright (C) 1997-2003 Andr Majorel et al
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//------------------------------------------------------------------------
//
//  Based on Yadex which incorporated code from DEU 5.21 that was put
//  in the public domain in 1994 by Raphal Quinet and Brendon Wyber.
//
//------------------------------------------------------------------------

#include "main.h"

#include "e_checks.h"
#include "m_dialog.h"
#include "editloop.h"
#include "editobj.h"
#include "e_cutpaste.h"
#include "r_misc.h"
#include "r_grid.h"
#include "e_linedef.h"
#include "e_sector.h"
#include "levels.h"
#include "objects.h"
#include "s_misc.h"
#include "selectn.h"
#include "selpath.h"
#include "x_mirror.h"
#include "x_hover.h"
#include "r_render.h"
#include "ui_window.h"


int active_when = 0;
int active_wmask = 0;

bool is_butl = false; //FIXME !!!!!


static int zoom_fit ();


Editor_State_c edit;


static const char *levelname;

static const Objid CANVAS (OBJ_NONE, OBJ_NO_CANVAS);


Editor_State_c::Editor_State_c()
    // FIXME !!!!
{
}

Editor_State_c::~Editor_State_c()
{
}


/*
 *  zoom_fit - adjust zoom factor to make level fit in window
 *
 *  Return 0 on success, non-zero on failure.
 */
static int zoom_fit ()
{
	if (NumVertices == 0)
		return 0;

	double xzoom;
	if (MapBound_hx - MapBound_lx)
		xzoom = .95 * ScrMaxX / (MapBound_hx - MapBound_lx);
	else
		xzoom = 1;
	double yzoom;
	if (MapBound_hy - MapBound_ly)
		yzoom = .9 * ScrMaxY / (MapBound_hy - MapBound_ly);
	else
		yzoom = 1;

	grid.NearestScale(MIN(xzoom, yzoom));

	grid.CenterMapAt((MapBound_lx + MapBound_hx) / 2, (MapBound_ly + MapBound_hy) / 2);
	return 0;
}


void UpdateHighlight(void)
{
	Objid obj;

	if (edit.pointer_in_window)
		GetCurObject(obj, edit.obj_type, edit.map_x, edit.map_y);
	else
		obj.nil();


	edit.highlighted = obj;

	int h_type = obj.type;

	if (obj.is_nil())
		h_type = edit.obj_type;

	if (obj())
		main_win->canvas->HighlightSet(edit.highlighted);
	else
		main_win->canvas->HighlightForget();

	// choose object to show in right-hand Panel
	int obj_idx = obj.num;
	int obj_count = (obj_idx < 0) ? 0 : 1;

	if (obj.is_nil() && edit.Selected->notempty())
	{
		selection_iterator_c it;

		edit.Selected->begin(&it);

		obj_idx = *it;
	}

	if (edit.Selected->notempty())
	{
		// handle object which is highlighted but NOT in the selection
		if (obj_idx >= 0 && ! edit.Selected->get(obj_idx))
			obj_count = 1;
		else
			obj_count = edit.Selected->count_obj();
	}

	switch (h_type)
	{
		case OBJ_THINGS:
			main_win->thing_box->SetObj(obj_idx, obj_count);
			break;

		case OBJ_LINEDEFS:
			main_win->line_box->SetObj(obj_idx, obj_count);
			break;

		case OBJ_SECTORS:
			main_win->sec_box->SetObj(obj_idx, obj_count);
			break;

		case OBJ_VERTICES:
			main_win->vert_box->SetObj(obj_idx, obj_count);
			break;

		default: break;
	}
}

bool GetCurrentObjects(selection_c *list)
{
	// returns false when there are no objects at all

	list->change_type(edit.obj_type);  // this also clears it

	if (edit.Selected->notempty())
	{
		list->merge(*edit.Selected);
		return true;
	}

	if (! edit.highlighted.is_nil())
	{
		list->set(edit.highlighted.num);
		return true;
	}

	return false;
}

void CMD_ChangeEditMode(char mode)
{
	obj_type_t prev_type = edit.obj_type;

	// Set the object type according to the new mode.
	switch (mode)
	{
		case 't': edit.obj_type = OBJ_THINGS;   break;
		case 'l': edit.obj_type = OBJ_LINEDEFS; break;
		case 's': edit.obj_type = OBJ_SECTORS;  break;
		case 'v': edit.obj_type = OBJ_VERTICES; break;
		case 'r': edit.obj_type = OBJ_RADTRIGS; break;

		default: FatalError("INTERNAL ERROR: unknown mode %d\n", mode);
	}

	edit.highlighted.nil();

	main_win->SetMode(mode);

	// convert selection (if necessary)

	// -AJA- Yadex (DEU?) would clear the selection if the mode didn't
	//       change.  TODO: consider an option to emulate that behavior.
	if (prev_type != edit.obj_type)
	{
		selection_c *prev_sel = edit.Selected;
		edit.Selected = new selection_c(edit.obj_type);

		ConvertSelection(prev_sel, edit.Selected);
		delete prev_sel;
	}

	UpdateHighlight();

	edit.RedrawMap = 1;
}

void CMD_SelectAll()
{
	int total = NumObjects(edit.obj_type);

	edit.Selected->change_type(edit.obj_type);
	edit.Selected->frob_range(0, total-1, BOP_ADD);
	edit.RedrawMap = 1;

	UpdateHighlight();
}

void CMD_UnselectAll()
{
	edit.Selected->change_type(edit.obj_type);
	edit.Selected->clear_all();
	edit.RedrawMap = 1;

	UpdateHighlight();
}


void CMD_Quit()
{
///??	edit.Selected->clear_all();

	main_win->action = UI_MainWin::QUIT;

	edit.RedrawMap = 1;
}



void EditorWheel(int delta, keymod_e mod)
{
    float S1 = grid.Scale;

    grid.AdjustScale((delta > 0) ? +1 : -1);
    grid.RefocusZoom(edit.map_x, edit.map_y, S1);
}


bool EditorKey(int is_key, keymod_e mod)
{
	// [Ctrl][L]: force redraw
	if (is_key == '\f')
	{
		edit.RedrawMap = 1;
	}

	// [ESC], [Q]: close
	else if (is_key == FL_Escape || is_key == 'Q')
	{
fprintf(stderr, "FL_Escape pressed\n");
		CMD_Quit();
	}

	else if (is_key == 'b')
	{
		main_win->ShowBrowser(2);
	}

#if 0 //!!!
	// [F5]: pop up the "Preferences" menu
	else if (is_key == FL_F+5)
	{
		Preferences (-1, -1);
		edit.RedrawMap = 1;
	}
#endif

#if 0 //!!!!!
	// [a]: pop up the "Set flag" menu
	else if (is_key == 'a'
			&& edit.menubar->highlighted () < 0
			&& (edit.Selected || edit.highlighted ()))
	{
		if (edit.obj_type == OBJ_LINEDEFS)
		{
			menu_linedef_flags->set_title ("Set linedef flag");
			e.modpopup->set (menu_linedef_flags, 0);
		}
		else if (edit.obj_type == OBJ_THINGS)
		{
			menu_thing_flags->set_title ("Set thing flag");
			e.modpopup->set (menu_thing_flags, 0);
		}
	}

	// [b]: pop up the "Toggle flag" menu
	else if (is_key == 'b'
			&& edit.menubar->highlighted () < 0
			&& (edit.Selected || edit.highlighted ()))
	{
		if (edit.obj_type == OBJ_LINEDEFS)
		{
			menu_linedef_flags->set_title ("Toggle linedef flag");
			e.modpopup->set (menu_linedef_flags, 0);
		}
		else if (edit.obj_type == OBJ_THINGS)
		{
			menu_thing_flags->set_title ("Toggle thing flag");
			e.modpopup->set (menu_thing_flags, 0);
		}
	}

	// [c]: pop up the "Clear flag" menu
	else if (is_key == 'c'
			&& edit.menubar->highlighted () < 0
			&& (edit.Selected || edit.highlighted ()))
	{
		if (edit.obj_type == OBJ_LINEDEFS)
		{
			menu_linedef_flags->set_title ("Clear linedef flag");
			e.modpopup->set (menu_linedef_flags, 0);
		}
		else if (edit.obj_type == OBJ_THINGS)
		{
			menu_thing_flags->set_title ("Clear thing flag");
			e.modpopup->set (menu_thing_flags, 0);
		}
	}
#endif

	// [M]: pop up the "Misc. operations" menu
	else if (is_key == 9876)
	{
		///---    edit.modpopup->set (edit.menubar->get_menu (MBI_MISC), 1);
	}

	// [F9]: pop up the "Insert a standard object" menu
	else if (is_key == FL_F+9)
	{
		///---    edit.modpopup->set (edit.menubar->get_menu (MBI_OBJECTS), 1);
	}

	// [F10]: pop up the "Checks" menu
	else if (is_key == FL_F+10)
	{
		CheckLevel (-1, -1);
		edit.RedrawMap = 1;
	}


	// [+], [=], wheel: zooming in
	else if (is_key == '+' || is_key == '=')
	{
		EditorWheel(+1, KM_none);
	}

	// [-], [_], wheel: zooming out
	else if (is_key == '-' || is_key == '_')
	{
		EditorWheel(-1, KM_none);
	}

	// [HOME]: centre window on centre of map
	// and set zoom to view the entire map
	else if (is_key == FL_Home)
	{
		CalculateLevelBounds();
		zoom_fit();
		edit.RedrawMap = 1;
	}

	// [1] - [9]: set the zoom factor
	else if (is_key >= '1' && is_key <= '9')
	{
		float S1 = grid.Scale;

		grid.ScaleFromDigit(is_key - '0');
		grid.RefocusZoom(edit.map_x, edit.map_y, S1);
	}

	// [Left], [Right], [Up], [Down]:
	// scroll <scroll_less> percents of a screenful.
	else if (is_key == FL_Left && grid.orig_x > -30000)
	{
		grid.orig_x -= (int) ((double) ScrMaxX * scroll_less / 100 / grid.Scale);
		edit.RedrawMap = 1;
	}
	else if (is_key == FL_Right && grid.orig_x < 30000)
	{
		grid.orig_x += (int) ((double) ScrMaxX * scroll_less / 100 / grid.Scale);
		edit.RedrawMap = 1;
	}
	else if (is_key == FL_Up && grid.orig_y < 30000)
	{
		grid.orig_y += (int) ((double) ScrMaxY * scroll_less / 100 / grid.Scale);
		edit.RedrawMap = 1;
	}
	else if (is_key == FL_Down && grid.orig_y > -30000)
	{
		grid.orig_y -= (int) ((double) ScrMaxY * scroll_less / 100 / grid.Scale);
		edit.RedrawMap = 1;
	}

	// [Pgup], [Pgdn], [Home], [End]:
	// scroll <scroll_more> percents of a screenful.
	else if (is_key == FL_Page_Up && (grid.orig_y) < /*MapBound_hy*/ 20000)
	{
		grid.orig_y += (int) ((double) ScrMaxY * scroll_more / 100 / grid.Scale);
		edit.RedrawMap = 1;
	}
	else if (is_key == FL_Page_Down && (grid.orig_y) > /*MapBound_ly*/ -20000)
	{
		grid.orig_y -= (int) ((double) ScrMaxY * scroll_more / 100 / grid.Scale);
		edit.RedrawMap = 1;
	}
#if 0
	else if (is_key == FL_Home && (grid.orig_x) > /*MapBound_lx*/ -20000)
	{
		grid.orig_x -= (int) ((double) ScrMaxX * scroll_more / 100 / grid.Scale);
		edit.RedrawMap = 1;
	}
	else if (is_key == FL_End && (grid.orig_x) < /*MapBound_hx*/ 20000)
	{
		grid.orig_x += (int) ((double) ScrMaxX * scroll_more / 100 / grid.Scale);
		edit.RedrawMap = 1;
	}
#endif

	// [l], [s], [t], [v], [r]: switch mode
	else if (is_key == 't' || is_key == 'v' || is_key == 'l' ||
	         is_key == 's' || is_key == 'r')
	{
		CMD_ChangeEditMode(is_key);
	}

#if 0
	// [e]: Select/unselect all linedefs in non-forked path
	else if (is_key == 'e' && edit.highlighted._is_linedef ())
	{
		edit.Selected->clear_all();
		
//!!!!!!		select_linedefs_path (&edit.Selected, edit.highlighted.num, BOP_ADD);
		edit.RedrawMap = 1;
	}

	// [Ctrl][e] Select/unselect all linedefs in path
	else if (is_key == '\5' && ! is_shift && edit.highlighted._is_linedef ())
	{
//!!!!!!		select_linedefs_path (&edit.Selected, edit.highlighted.num, BOP_TOGGLE);
		edit.RedrawMap = 1;
	}
	// [E]: add linedef and split sector -- [AJA]
	else if (is_key == 'E' && edit.obj_type == OBJ_VERTICES)
	{
		if (edit.Selected)
		{
///!!!!!		MiscOperations (edit.obj_type, &edit.Selected, 5);
			edit.RedrawMap = 1;
		}
	}
	// [E]: Select/unselect all 1s linedefs in path
	else if (is_key == 'E' && edit.highlighted._is_linedef ())
	{
		edit.Selected->clear_all();
//!!!!!!		select_1s_linedefs_path (&edit.Selected, edit.highlighted.num, BOP_ADD);
		edit.RedrawMap = 1;
	}

	// [Ctrl][Shift][e]: Select/unselect all 1s linedefs in path
	else if (is_key == '\5' && is_shift && edit.highlighted._is_linedef ())
	{
//!!!!!!		select_1s_linedefs_path (&edit.Selected, edit.highlighted.num, BOP_TOGGLE);
		edit.RedrawMap = 1;
	}
#endif

	// [G]: to increase the grid step
	else if (is_key == 'G')
	{
		grid.AdjustStep(+1);
	}

	// [g]: decrease the grid step
	else if (is_key == 'g')
	{
		grid.AdjustStep(-1);
	}

	// [h]: display or hide the grid
	else if (is_key == 'h')
	{
		grid.shown = ! grid.shown;
		edit.RedrawMap = 1;
	}

	//???  // [H]: reset the grid to grid_step_max
	//???  else if (is_key == 'H')
	//???  {
	//???    e.grid_step = e.grid_step_max;
	//???    edit.RedrawMap = 1;
	//???  }

	// [f]: toggle the snap_to_grid flag
	else if (is_key == 'f')
	{
		grid.locked = false;
		grid.snap   = ! grid.snap;

		main_win->info_bar->UpdateLock();
	}

	// [k]: toggle the lock_grip_step flag
	else if (is_key == 'k')
	{
		grid.locked = ! grid.locked;
		grid.snap   = true;

		main_win->info_bar->UpdateLock();
	}

	// [q]: quantize objects to the grid
	else if (is_key == 'q')
	{
		CMD_QuantizeObjects();
	}

	// [n], [>]: highlight the next object
	else if ((is_key == 'n' )
			&& ( edit.highlighted ()))
	{
		int t = edit.highlighted () ? edit.highlighted.type : edit.obj_type;
		obj_no_t nmax = GetMaxObjectNum (t);
		if (is_obj (nmax))
		{
			if (edit.highlighted.is_nil ())
			{
				edit.highlighted.type = (obj_type_t)t;
				edit.highlighted.num = 0;
			}
			else
			{
				edit.highlighted.num++;
				if (edit.highlighted.num > nmax)
					edit.highlighted.num = 0;
			}
			GoToObject (edit.highlighted);
			edit.RedrawMap = 1;
		}
	}

	// [p], [<]: highlight the previous object
	else if ((is_key == 'p' )
			&& ( edit.highlighted ()))
	{
		int t = edit.highlighted () ? edit.highlighted.type : edit.obj_type;
		obj_no_t nmax = GetMaxObjectNum (t);
		if (is_obj (nmax))
		{
			if (edit.highlighted.is_nil ())
			{
				edit.highlighted.type = (obj_type_t)t;
				edit.highlighted.num = nmax;
			}
			else
			{
				edit.highlighted.num--;
				if (edit.highlighted.num < 0)
					edit.highlighted.num = nmax;
			}
			GoToObject (edit.highlighted);
			edit.RedrawMap = 1;
		}
	}

	// [j], [#]: jump to object by number
	else if ((is_key == 'j' || is_key == '#')
			&& ( edit.highlighted ()))
	{
		const char *buf = fl_input("Enter index", "");

		if (buf)
		{
			Objid target_obj;
			target_obj.type = edit.obj_type;
			target_obj.num  = atoi(buf);

			if (target_obj ())
			{
				GoToObject (target_obj);
				edit.RedrawMap = 1;
			}
		}
	}

	// [f]: find object by type
#if 0
	else if (is_key == 'f' && ( edit.highlighted ())) 
	{
		Objid find_obj;
		int otype;
		obj_no_t omax,onum;
		find_obj.type = edit.highlighted () ? edit.highlighted.type : edit.obj_type;
		onum = find_obj.num  = edit.highlighted () ? edit.highlighted.num  : 0;
		omax = GetMaxObjectNum(find_obj.type);
		switch (find_obj.type)
		{
			case OBJ_SECTORS:
				if ( ! InputSectorType( 84, 21, &otype))
				{
					for (onum = edit.highlighted () ? onum + 1 : onum; onum <= omax; onum++)
						if (Sectors[onum]->type == otype)
						{
							find_obj.num = onum;
							GoToObject(find_obj);
							break;
						}
				}
				break;
			case OBJ_THINGS:
				if ( ! InputThingType( 42, 21, &otype))
				{
					for (onum = edit.highlighted () ? onum + 1 : onum; onum <= omax; onum++)
						if (Things[onum]->type == (wad_ttype_t) otype)
						{
							find_obj.num = onum;
							GoToObject(find_obj);
							break;
						}
				}
				break;
			case OBJ_LINEDEFS:
				if ( ! InputLinedefType( 0, 21, &otype))
				{
					for (onum = edit.highlighted () ? onum + 1 : onum; onum <= omax; onum++)
						if (LineDefs[onum]->type == otype)
						{
							find_obj.num = onum;
							GoToObject(find_obj);
							break;
						}
				}
				break;
		}
		edit.RedrawMap = 1;
	}
#endif

	// SPACE: clear selection and redraw the map
	else if (is_key == ' ')
	{
		CMD_UnselectAll();
	}

	// [o]: copy a group of objects
	else if (is_key == 'o'
			&& (edit.Selected || edit.highlighted ()))
	{
		if (CMD_Copy())
		{
			CMD_Paste();
		}
	}

	else if (is_key == 'W')
	{
		CMD_RotateObjects(true);
	}
	else if (is_key == 'X')
	{
		CMD_RotateObjects(false);
	}

	else if (is_key == 'H')
	{
		CMD_MirrorObjects(false);
	}
	else if (is_key == 'I')
	{
		CMD_MirrorObjects(true);
	}

	else if (is_key == 'm')
	{
		CMD_ScaleObjects(true);
	}
	else if (is_key == 'M')
	{
		CMD_ScaleObjects(false);
	}

	// 'w': spin things 1/8 turn counter-clockwise
	else if (is_key == 'w' && edit.obj_type == OBJ_THINGS
			&& (edit.Selected || edit.highlighted ()))
	{
		CMD_SpinThings(+45);
	}

	// 'x': spin things 1/8 turn clockwise
	else if (is_key == 'x' && edit.obj_type == OBJ_THINGS)
	{
		CMD_SpinThings(-45);
	}

	// '[': lower floor height
	else if ((is_key == '[' || is_key == '{') && edit.obj_type == OBJ_SECTORS)
	{
		CMD_MoveFloors((is_key == '[') ? -16 : -2);
	}
	// ']': raise floor height
	else if ((is_key == ']' || is_key == '}') && edit.obj_type == OBJ_SECTORS)
	{
		CMD_MoveFloors((is_key == ']') ? +16 : +2);
	}

	// '<': lower ceiling height
	else if ((is_key == ',' || is_key == '<') && edit.obj_type == OBJ_SECTORS)
	{
		CMD_MoveCeilings((is_key == ',') ? -16 : -2);
	}
	// '>': raise ceiling height
	else if ((is_key == '.' || is_key == '>') && edit.obj_type == OBJ_SECTORS)
	{
		CMD_MoveCeilings((is_key == '.') ? +16 : +2);
	}


	// [w]: split linedefs and sectors
//!!!	else if (is_key == 'w' && edit.obj_type == OBJ_LINEDEFS
//!!!			&& edit.Selected && edit.Selected->next && ! edit.Selected->next->next)
//!!!	{
//!!!		SplitLineDefsAndSector (edit.Selected->next->objnum, edit.Selected->objnum);
//!!!		edit.Selected->clear_all();
//!!!		edit.RedrawMap = 1;
//!!!	}

//!!!	// [w]: split sector between vertices
//!!!	else if (is_key == 'w' && edit.obj_type == OBJ_VERTICES
//!!!			&& edit.Selected && edit.Selected->next && ! edit.Selected->next->next)
//!!!	{
//!!!		SplitSector (edit.Selected->next->objnum, edit.Selected->objnum);
//!!!		edit.Selected->clear_all();
//!!!		edit.RedrawMap = 1;
//!!!	}


	// [x]: split linedefs
	else if (is_key == 'x' && edit.obj_type == OBJ_LINEDEFS)
	{
		CMD_SplitLineDefs();
	}

	// [w]: flip linedefs
	else if (is_key == 'w' && edit.obj_type == OBJ_LINEDEFS)
	{
		CMD_FlipLineDefs();
	}

	// [Ctrl][x]: exchange objects numbers
//!!!	else if (is_key == 24)
//!!!	{
//!!!		if (! edit.Selected
//!!!				|| ! edit.Selected->next
//!!!				|| (edit.Selected->next)->next)
//!!!		{
//!!!			Beep ();
//!!!			Notify (-1, -1, "You must select exactly two objects", 0);
//!!!			edit.RedrawMap = 1;
//!!!		}
//!!!		else
//!!!		{
//!!!			exchange_objects_numbers (edit.obj_type, edit.Selected, true);
//!!!			edit.RedrawMap = 1;
//!!!		}
//!!!	}

	// [Ctrl][k]: cut a slice out of a sector
//!!!	else if (is_key == 11 && edit.obj_type == OBJ_LINEDEFS
//!!!			&& edit.Selected && edit.Selected->next && ! edit.Selected->next->next)
//!!!	{
//!!!		sector_slice (edit.Selected->next->objnum, edit.Selected->objnum);
//!!!		edit.Selected->clear_all();
//!!!		edit.RedrawMap = 1;
//!!!	}

	// [Del]: delete the current object
	else if ((is_key == '\b' || is_key == FL_BackSpace || is_key == FL_Delete))
	{
		bool keep_unused = (mod == KM_SHIFT) ? true : false;

		CMD_Delete(false, keep_unused);
	}

	// [Ins]: insert a new object
	else if (is_key == 'a' || is_key == 'A' || is_key == FL_Insert)
	{
		InsertNewObject();

		UpdateHighlight();

		edit.RedrawMap = 1;

	}

	// [Z] Set sector on surrounding linedefs (AJA)
//!!!!!!	else if (is_key == 'Z' && edit.pointer_in_window) 
//!!!!!!	{
//!!!!!!		if (edit.obj_type == OBJ_SECTORS && edit.Selected)
//!!!!!!		{
//!!!!!!			SuperSectorSelector (edit.map_x, edit.map_y,
//!!!!!!					edit.Selected->objnum);
//!!!!!!		}
//!!!!!!		else
//!!!!!!		{
//!!!!!!			SuperSectorSelector (edit.map_x, edit.map_y, OBJ_NO_NONE);
//!!!!!!		}
//!!!!!!		edit.RedrawMap = 1;
//!!!!!!	}

	// [K] limit shown things to specific skill (AJA)
	else if (is_key == 'K' && edit.obj_type == OBJ_THINGS)
	{
		active_wmask ^= 1;
		active_when = active_wmask;
		edit.RedrawMap = 1;
	}

	// [!] Debug info (not documented)
	else if (is_key == '!')
	{
		DumpSelection (edit.Selected);
	}


	// [T] Transfer properties to selected objects (AJA)
	else if (is_key == 'T' && edit.Selected 
			&& edit.highlighted.num >= 0)
	{
#if 0
		switch (edit.obj_type)
		{
			case OBJ_SECTORS:
				TransferSectorProperties (edit.highlighted.num, edit.Selected);
				edit.RedrawMap = 1;
				break;
			case OBJ_THINGS:
				TransferThingProperties (edit.highlighted.num, edit.Selected);
				edit.RedrawMap = 1;
				break;
			case OBJ_LINEDEFS:
				TransferLinedefProperties (edit.highlighted.num, edit.Selected);
				edit.RedrawMap = 1;
				break;
			default:
				Beep ();
				break;
		}
#endif
	}


	// [&] Show object numbers
	else if (is_key == '&')
	{
		edit.show_object_numbers = ! edit.show_object_numbers;
		edit.RedrawMap = 1;
	}

	// [%] Show things sprites
	else if (is_key == '%')
	{
		edit.show_things_sprites = ! edit.show_things_sprites;
		edit.show_things_squares = ! edit.show_things_sprites;  // Not a typo !
		edit.RedrawMap = 1;
	}

	// [PRTSCR]: save a screen shot.   FIXME
	else if (is_key == FL_Print)
	{
		Beep();
	}

	else
	{
		// fprintf(stderr, "UNKNOWN KEY: %d (0x%04x)\n", is_key, is_key);
		return false;
	}

	return true;
}


void EditorMousePress(keymod_e mod)
{
	bool is_ctrl = (mod == KM_CTRL);

	is_butl = true;

	Objid object;      // The object under the pointer

	GetCurObject(object, edit.obj_type, edit.map_x, edit.map_y);

	/* Clicking on an empty space starts a new selection box.
	   Unless [Ctrl] is pressed, it also clears the current selection. */
	if (true
			&& object.is_nil ())  //!!!!!!! FIXME
	{
		edit.clicked    = CANVAS;
		edit.click_ctrl = is_ctrl;

		if (! is_ctrl)
		{
			edit.Selected->clear_all();
			edit.RedrawMap = 1;
		}

		main_win->canvas->SelboxBegin(edit.map_x, edit.map_y);

		///---    edit.selbox->set_1st_corner ;
		///---    edit.selbox->set_2nd_corner (edit.map_x, edit.map_y);
		///---  main_win->canvas->redraw();
		return;
	}

	/* Clicking on an unselected object unselects
	   everything but that object. Additionally,
	   we write the number of the object in case
	   the user is about to drag it. */
	if (! is_ctrl
			&& ! IsSelected (edit.Selected, object.num))
	{
		edit.clicked        = object;
		edit.click_ctrl     = 0;
		edit.click_time     = 0; // is.time;

		//@@@@@
		// if (edit.Selected == NOTHING)

		edit.Selected->clear_all();
		SelectObject (edit.Selected, object.num);

//????		/* I don't like having to do that */
//????		if (object.type == OBJ_THINGS && object ())
//????			MoveObjectsToCoords (object.type, 0,
//????					Things[object.num]->x, Things[object.num]->y, 0);
//????		else if (object.type == OBJ_VERTICES && object ())
//????			MoveObjectsToCoords (object.type, 0,
//????					Vertices[object.num]->x, Vertices[object.num]->y, 0);
//????		else
//????			MoveObjectsToCoords (object.type, 0,
//????					edit.map_x, edit.map_y, grid.snap ? grid.step : 0);

		edit.RedrawMap = 1;

		main_win->canvas->redraw();
		return;
	}

#if 0  /* Second click of a double click on an object */
	if (! is_ctrl
			&& IsSelected (edit.Selected, object.num)
			&& object  == edit.clicked
			&& is.time - edit.click_time <= (unsigned long) double_click_timeout)
	{
		// Very important! If you don't do that, the release of the
		// click that closed the properties menu will drag the object.
		edit.clicked.nil ();
		send_event (YK_RETURN);

		main_win->canvas->redraw();
		return;
	}
#endif

	/* Clicking on a selected object does nothing ;
	   the user might want to drag the selection. */
	if (! is_ctrl
			&& IsSelected (edit.Selected, object.num))
	{
		edit.clicked        = object;
		edit.click_ctrl     = 0;
		edit.click_time     = 0; /// is.time;

///????		/* I don't like having to do that */
///????		if (object.type == OBJ_THINGS && object ())
///????			MoveObjectsToCoords (object.type, 0,
///????					Things[object.num]->x, Things[object.num]->y, 0);
///????		else if (object.type == OBJ_VERTICES && object ())
///????			MoveObjectsToCoords (object.type, 0,
///????					Vertices[object.num]->x, Vertices[object.num]->y, 0);
///????		else
///????			MoveObjectsToCoords (object.type, 0,
///????					edit.map_x, edit.map_y, grid.snap ? grid.step : 0);

		edit.RedrawMap = 1;
		return;
	}

	/* Clicking on selected object with [Ctrl] pressed unselects it.
	   Clicking on unselected object with [Ctrl] pressed selects it. */
	if (is_ctrl
			&& object ())
	{
		edit.clicked        = object;
		edit.click_ctrl     = 1;

		if (IsSelected (edit.Selected, object.num))
			UnSelectObject (edit.Selected, object.num);
		else
			SelectObject (edit.Selected, object.num);

		edit.RedrawMap = 1;

		return;
	}

}

void EditorMouseRelease()
{
	is_butl = false;

	/* Releasing the button while there was a selection box
	   causes all the objects within the box to be selected. */
	if (main_win->canvas->isSelboxActive())
	{
		int x1, y1, x2, y2;
		main_win->canvas->SelboxFinish(&x1, &y1, &x2, &y2);

		SelectObjectsInBox(edit.Selected, edit.obj_type, x1, y1, x2, y2);

		UpdateHighlight();

		edit.RedrawMap = 1;
		return;
	}

	/* Releasing the button while dragging : drop the selection. */
	// FIXME : should call this automatically when switching tool
	if (main_win->canvas->isDragActive())
	{
		int dx, dy;
		main_win->canvas->DragFinish(&dx, &dy);

		if (! (dx==0 && dy==0))
		{
			CMD_MoveObjects(dx, dy);
		}

		edit.RedrawMap = 1;
		return;
	}

}

void EditorLeaveWindow(void)
{
	edit.pointer_in_window = false;
	
	UpdateHighlight();
}

void EditorMouseMotion(int x, int y, int map_x, int map_y, bool drag)
{
	edit.map_x = map_x;
	edit.map_y = map_y;
	edit.pointer_in_window = true; // FIXME

	if (edit.pointer_in_window)
		main_win->info_bar->SetMouse(edit.map_x, edit.map_y);

	// fprintf(stderr, "MOUSE MOTION: %d,%d  map: %d,%d\n", x, y, edit.map_x, edit.map_y);

	if (! drag)
	{
		UpdateHighlight();
	}
	/* Moving the pointer with the left button pressed
	   and a selection box exists : move the second
	   corner of the selection box. */
	else if (true
			&& is_butl  /* FIXME: edit.selecting */
			&& edit.clicked == CANVAS)
	{
		main_win->canvas->SelboxUpdate(edit.map_x, edit.map_y);
		return;
	}

	/* Moving the pointer with the left button pressed
	   but no selection box exists and [Ctrl] was not
	   pressed when the button was pressed :
	   drag the selection. */
	if (main_win->canvas->isDragActive())
	{
		main_win->canvas->DragUpdate(edit.map_x, edit.map_y);
	}
	else
	if (true
		&& is_butl  /* FIXME: edit.dragging */
		&& edit.clicked ()
		&& ! edit.click_ctrl
		&& edit.Selected->notempty())
	{
		main_win->canvas->DragBegin(edit.map_x, edit.map_y);

//!!!			if (MoveObjectsToCoords (edit.clicked.type, edit.Selected,
//!!!						edit.map_x, edit.map_y, grid.snap ? grid.step : 0))
//!!!				edit.RedrawMap = 1;

		return;
	}
}


void EditorResize(int is_width, int is_height)
{
	edit.RedrawMap = 1;
}


void EditorAutoScroll()
{
  // Auto-scrolling: scroll the map automatically
  // when the mouse pointer rests near the edge
  // of the window.
  // Scrolling is disabled when a pull-down menu
  // is visible because it would be annoying to see
  // the map scrolling while you're searching
  // through the menus.

#if 0  // DISABLED FOR TIME BEING

  if (is.in_window
      && autoscroll
      && ! is.scroll_lock
      && edit.menubar->pulled_down () < 0)
  {
    unsigned distance;    // In pixels

#define actual_move(total,dist) \
    ((int) (((total * autoscroll_amp / 100)\
        * ((double) (autoscroll_edge - dist) / autoscroll_edge))\
      / Scale))

    distance = is.y;
    // The reason for the second member of the condition
    // is that we don't want to scroll when the user is
    // simply reaching for a menu...
    if (distance <= autoscroll_edge
        && edit.menubar->is_under_menubar_item (is.x) < 0)
    {
      if ((grid.orig_y) < /*MapBound_hy*/ 20000)
      {
        grid.orig_y += actual_move (ScrMaxY, distance);
        edit.RedrawMap = 1;
      }
    }

    distance = ScrMaxY - is.y;
    if (distance <= autoscroll_edge)
    {
      if ((grid.orig_y) > /*MapBound_ly*/ -20000)
      {
        grid.orig_y -= actual_move (ScrMaxY, distance);
        edit.RedrawMap = 1;
      }
    }

    distance = is.x;
    if (distance <= autoscroll_edge)
    {
      if ((grid.orig_x) > /*MapBound_lx*/ -20000)
      {
        grid.orig_x -= actual_move (ScrMaxX, distance);
        edit.RedrawMap = 1;
      }
    }

    // The reason for the second member of the condition
    // is that we don't want to scroll when the user is
    // simply reaching for the "Help" menu...
    // Note: the ordinate "3 * FONTH" is of course not
    // critical. It's just a rough approximation.
    distance = ScrMaxX - is.x;
    if (distance <= autoscroll_edge && (unsigned) is.y >= 3 * FONTH)
    {
      if ((grid.orig_x) < /*MapBound_hx*/ 20000)
      {
        grid.orig_x += actual_move (ScrMaxX, distance);
        edit.RedrawMap = 1;
      }
    }
  }
#endif
}


/*
  the editor main loop
*/

void EditorLoop (const char *_levelname)
{
    levelname = _levelname;

    memset (&edit, 0, sizeof(edit));  /* Catch-all */

    edit.move_speed          = 20;
    edit.extra_zoom          = 0;
    edit.obj_type            = OBJ_THINGS;

    edit.show_object_numbers = false;
    edit.show_things_squares = false;
    edit.show_things_sprites = true;

    edit.clicked.nil ();
    //edit.click_obj_no        = OBJ_NO_NONE;
    //edit.click_obj_type      = -1;
    edit.click_ctrl          = 0;
    edit.highlighted.nil ();
    //edit.highlight_obj_no    = OBJ_NO_NONE;
    //edit.highlight_obj_type  = -1;

    edit.Selected = new selection_c(edit.obj_type);

    MadeChanges = 0;


	grid.Init();

	CalculateLevelBounds();
    zoom_fit();

    Render3D_Init();


    main_win->canvas->set_edit_info(&edit);

	main_win->SetMode('t');

	UpdateHighlight();


    edit.RedrawMap = 1;

    for (;;)
    {
		if (main_win->action == UI_MainWin::QUIT)
		{
			if (! MadeChanges ||
				Confirm (-1, -1, "You have unsaved changes."
						" Do you really want to quit?", 0))
			{
				break;
			}

			main_win->action = 0;
		}

        Fl::wait(0.2);

        if (edit.RedrawMap)
        {
            main_win->canvas->redraw();

            edit.RedrawMap = 0;
        }
    }
}

//--- editor settings ---
// vi:ts=4:sw=4:noexpandtab
