//------------------------------------------------------------------------
//  MIRROR / ROTATE / ETC OPS
//------------------------------------------------------------------------
//
//  Eureka DOOM Editor
//
//  Copyright (C) 2001-2009 Andrew Apted
//  Copyright (C) 1997-2003 Andr Majorel et al
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//------------------------------------------------------------------------
//
//  Based on Yadex which incorporated code from DEU 5.21 that was put
//  in the public domain in 1994 by Raphal Quinet and Brendon Wyber.
//
//------------------------------------------------------------------------

#include "main.h"

#include <map>

#include "m_bitvec.h"
#include "e_linedef.h"
#include "e_sector.h"
#include "e_things.h"
#include "e_vertex.h"
#include "editloop.h"
#include "levels.h"
#include "m_game.h"
#include "r_grid.h"
#include "selectn.h"
#include "x_mirror.h"


/*
 *  Return the coordinates of the centre of a group of objects
 */
void centre_of_objects (selection_c * list, int *x, int *y)
{
	*x = *y = 0;

	if (list->empty())
		return;

	selection_iterator_c it;

	double sum_x = 0;
	double sum_y = 0;

	int count = 0;

	switch (list->what_type())
	{
		case OBJ_THINGS:
		{
			for (list->begin(&it); !it.at_end(); ++it, ++count)
			{
				sum_x += Things[*it]->x;
				sum_y += Things[*it]->y;
			}
			break;
		}

		case OBJ_VERTICES:
		{
			for (list->begin(&it); !it.at_end(); ++it, ++count)
			{
				sum_x += Vertices[*it]->x;
				sum_y += Vertices[*it]->y;
			}
			break;
		}

		case OBJ_RADTRIGS:
		{
			for (list->begin(&it); !it.at_end(); ++it, ++count)
			{
				sum_x += RadTrigs[*it]->mx;
				sum_y += RadTrigs[*it]->my;
			}
			break;
		}

		// everything else: just use the vertices
		default:
		{
			selection_c verts(OBJ_VERTICES);
			ConvertSelection(list, &verts);
			centre_of_objects(&verts, x, y);
			return;
		}
	}

	SYS_ASSERT(count > 0);

	*x = I_ROUND(sum_x / count);
	*y = I_ROUND(sum_y / count);
}


/*
 *  Returns a bounding box that completely includes a group of objects
 */
void bbox_of_objects(selection_c * list, int *x1, int *y1, int *x2, int *y2)
{
	if (list->empty())
	{
		*x1 = *y1 = 0;
		*x2 = *y2 = 0;
		return;
	}

	*x1 = *y1 = +777777;
	*x2 = *y2 = -777777;

	selection_iterator_c it;

	switch (list->what_type())
	{
		case OBJ_THINGS:
		{
			for (list->begin(&it); !it.at_end(); ++it)
			{
				const Thing *T = Things[*it];

				const thingtype_t *info = M_GetThingType(T->type);
				int r = info->radius;

				if (T->x - r < *x1) *x1 = T->x - r;
				if (T->y - r < *y1) *y1 = T->y - r;
				if (T->x + r > *x2) *x2 = T->x + r;
				if (T->y + r > *y2) *y2 = T->y + r;
			}
			break;
		}

		case OBJ_VERTICES:
		{
			for (list->begin(&it); !it.at_end(); ++it)
			{
				const Vertex *V = Vertices[*it];

				if (V->x < *x1) *x1 = V->x;
				if (V->y < *y1) *y1 = V->y;
				if (V->x > *x2) *x2 = V->x;
				if (V->y > *y2) *y2 = V->y;
			}
			break;
		}

		case OBJ_RADTRIGS:
		{
			for (list->begin(&it); !it.at_end(); ++it)
			{
				const RadTrig *R = RadTrigs[*it];

				int rx = (R->rw+1) / 2;
				int ry = (R->rh+1) / 2;

				if (R->mx - rx < *x1) *x1 = R->mx - rx;
				if (R->my - ry < *y1) *y1 = R->my - ry;
				if (R->mx + rx > *x2) *x2 = R->mx + rx;
				if (R->my + ry > *y2) *y2 = R->my + ry;
			}
			break;
		}

		// everything else: just use the vertices
		default:
		{
			selection_c verts(OBJ_VERTICES);
			ConvertSelection(list, &verts);
			bbox_of_objects(&verts, x1, y1, x2, y2);
			return;
		}
	}

	SYS_ASSERT(*x1 <= *x2);
	SYS_ASSERT(*y1 <= *y2);
}


static void DoMirrorThings(selection_c *list, bool is_vert, int cx, int cy)
{
	selection_iterator_c it;

	for (list->begin(&it); !it.at_end(); ++it)
	{
		const Thing * T = Things[*it];

		if (is_vert)
		{
			BA_ChangeTH(*it, Thing::F_Y, 2*cy - T->y);

			if (T->angle != 0)
				BA_ChangeTH(*it, Thing::F_ANGLE, 360 - T->angle);
		}
		else
		{
			BA_ChangeTH(*it, Thing::F_X, 2*cx - T->x);

			if (T->angle > 180)
				BA_ChangeTH(*it, Thing::F_ANGLE, 540 - T->angle);
			else
				BA_ChangeTH(*it, Thing::F_ANGLE, 180 - T->angle);
		}
	}
}


void CMD_MirrorObjects(bool is_vert)
{
	selection_c list;
	selection_iterator_c it;

	if (! GetCurrentObjects(&list) || edit.obj_type == OBJ_RADTRIGS)
	{
		Beep();
		return;
	}

	int cx, cy;
	centre_of_objects(&list, &cx, &cy);
	
	BA_Begin();

	if (edit.obj_type == OBJ_THINGS)
	{
		DoMirrorThings(&list, is_vert, cx, cy);
	}
	else
	{
		// handle things in Sectors mode too
		if (edit.obj_type == OBJ_SECTORS)
		{
			selection_c things(OBJ_THINGS);
			ConvertSelection(&list, &things);

			DoMirrorThings(&things, is_vert, cx, cy);
		}

		// everything else just modifies the vertices
		selection_c verts(OBJ_VERTICES);
		ConvertSelection(&list, &verts);

		for (verts.begin(&it); !it.at_end(); ++it)
		{
			const Vertex * V = Vertices[*it];

			if (is_vert)
				BA_ChangeVT(*it, Vertex::F_Y, 2*cy - V->y);
			else
				BA_ChangeVT(*it, Vertex::F_X, 2*cx - V->x);
		}

		// flip linedefs too !!
		selection_c lines(OBJ_LINEDEFS);
		ConvertSelection(&verts, &lines);

		for (lines.begin(&it); !it.at_end(); ++it)
		{
			LineDef * L = LineDefs[*it];

			int start = L->start;
			int end   = L->end;

			BA_ChangeLD(*it, LineDef::F_START, end);
			BA_ChangeLD(*it, LineDef::F_END, start);
		}
	}

	BA_End();
}


static void DoRotate90Things(selection_c *list, bool anti_clockwise, int cx, int cy)
{
	selection_iterator_c it;

	for (list->begin(&it); !it.at_end(); ++it)
	{
		const Thing * T = Things[*it];

		int old_x = T->x;
		int old_y = T->y;

		if (anti_clockwise)
		{
			BA_ChangeTH(*it, Thing::F_X, cx - old_y + cy);
			BA_ChangeTH(*it, Thing::F_Y, cy + old_x - cx);

			BA_ChangeTH(*it, Thing::F_ANGLE, calc_new_angle(T->angle, +90));
		}
		else
		{
			BA_ChangeTH(*it, Thing::F_X, cx + old_y - cy);
			BA_ChangeTH(*it, Thing::F_Y, cy - old_x + cx);

			BA_ChangeTH(*it, Thing::F_ANGLE, calc_new_angle(T->angle, -90));
		}
	}
}



void CMD_RotateObjects(bool anti_clockwise)  // 90 degrees
{
	selection_c list;
	selection_iterator_c it;

	if (! GetCurrentObjects(&list) || edit.obj_type == OBJ_RADTRIGS)
	{
		Beep();
		return;
	}

	int cx, cy;
	centre_of_objects(&list, &cx, &cy);
	
	BA_Begin();

	if (edit.obj_type == OBJ_THINGS)
	{
		DoRotate90Things(&list, anti_clockwise, cx, cy);
	}
	else
	{
		// handle things inside sectors
		if (edit.obj_type == OBJ_SECTORS)
		{
			selection_c things(OBJ_THINGS);
			ConvertSelection(&list, &things);

			DoRotate90Things(&things, anti_clockwise, cx, cy);
		}

		// everything else just rotates the vertices
		selection_c verts(OBJ_VERTICES);
		ConvertSelection(&list, &verts);

		for (verts.begin(&it); !it.at_end(); ++it)
		{
			const Vertex * V = Vertices[*it];

			int old_x = V->x;
			int old_y = V->y;

			if (anti_clockwise)
			{
				BA_ChangeVT(*it, Vertex::F_X, cx - old_y + cy);
				BA_ChangeVT(*it, Vertex::F_Y, cy + old_x - cx);
			}
			else
			{
				BA_ChangeVT(*it, Vertex::F_X, cx + old_y - cy);
				BA_ChangeVT(*it, Vertex::F_Y, cy - old_x + cx);
			}
		}
	}

	BA_End();
}
 

static void DoScale2xThings(selection_c *list, bool is_half, int cx, int cy)
{
	selection_iterator_c it;

	for (list->begin(&it); !it.at_end(); ++it)
	{
		const Thing * T = Things[*it];

		int dx = T->x - cx;
		int dy = T->y - cy;

		if (is_half)
		{
			BA_ChangeTH(*it, Thing::F_X, cx + dx / 2);
			BA_ChangeTH(*it, Thing::F_Y, cy + dy / 2);
		}
		else
		{
			BA_ChangeTH(*it, Thing::F_X, cx + dx * 2);
			BA_ChangeTH(*it, Thing::F_Y, cy + dy * 2);
		}
	}
}


void CMD_ScaleObjects(bool is_half)
{
	selection_c list;
	selection_iterator_c it;

	if (! GetCurrentObjects(&list))
	{
		Beep();
		return;
	}

	int cx, cy, hx, hy;
	if (true)  // FIXME: CONFIG OPTION
		bbox_of_objects(&list, &cx, &cy, &hx, &hy);
	else
		centre_of_objects(&list, &cx, &cy);

	BA_Begin();

	if (edit.obj_type == OBJ_RADTRIGS)
	{
		// Note: the positions of the trigger(s) are not changed
		for (list.begin(&it); !it.at_end(); ++it)
		{
			const RadTrig * R = RadTrigs[*it];

			int new_rw = is_half ? ((R->rw+1) / 2) : (R->rw * 2);
			int new_rh = is_half ? ((R->rh+1) / 2) : (R->rh * 2);

			BA_ChangeRAD(*it, RadTrig::F_RW, new_rw);
			BA_ChangeRAD(*it, RadTrig::F_RH, new_rh);
		}
	}
	else if (edit.obj_type == OBJ_THINGS)
	{
		DoScale2xThings(&list, is_half, cx, cy);
	}
	else
	{
		// handle things inside sectors
		if (edit.obj_type == OBJ_SECTORS)
		{
			selection_c things(OBJ_THINGS);
			ConvertSelection(&list, &things);

			DoScale2xThings(&things, is_half, cx, cy);
		}

		// everything else just scales the vertices
		selection_c verts(OBJ_VERTICES);
		ConvertSelection(&list, &verts);

		for (verts.begin(&it); !it.at_end(); ++it)
		{
			const Vertex * V = Vertices[*it];

			int dx = V->x - cx;
			int dy = V->y - cy;

			if (is_half)
			{
				BA_ChangeVT(*it, Vertex::F_X, cx + dx / 2);
				BA_ChangeVT(*it, Vertex::F_Y, cy + dy / 2);
			}
			else
			{
				BA_ChangeVT(*it, Vertex::F_X, cx + dx * 2);
				BA_ChangeVT(*it, Vertex::F_Y, cy + dy * 2);
			}
		}
	}

	BA_End();
}


#if 0
static void move_coord_for_quant(int *nx, int *ny, int dx, int dy, int pass)
{
	if (dx < 0) *nx -= pass * grid.step / 3;
	if (dx > 0) *nx += pass * grid.step / 3;

	if (dy < 0) *ny -= pass * grid.step / 3;
	if (dy > 0) *ny += pass * grid.step / 3;
}
#endif


static void DoQuantizeVertices(selection_c *verts)
{
	std::map<int, int> old_x;
	std::map<int, int> old_y;

	selection_iterator_c it;

	for (verts->begin(&it); !it.at_end(); ++it)
	{
		const Vertex * V = Vertices[*it];

		old_x[*it] = V->x;
		old_y[*it] = V->y;

		BA_ChangeVT(*it, Vertex::F_X, grid.ForceSnapX(V->x));
		BA_ChangeVT(*it, Vertex::F_Y, grid.ForceSnapY(V->y));
	}

	// check for linedefs which have become zero-length, and
	// "fix" them by moving the vertex somewhere else.

	selection_c lines(OBJ_LINEDEFS);

	for (int n = 0; n < NumLineDefs; n++)
		if (verts->get(LineDefs[n]->start) || verts->get(LineDefs[n]->end))
			lines.set(n);

	if (lines.empty())
		return;

#if 0  // FIXME: PREVENTING "BAD" QUANTIZATIONS IS HARD!!

	for (int pass = 1; pass < 20; pass++)
	{
		int count = 0;
		selection_iterator_c it;

		for (lines.begin(&it); !it.at_end(); ++it)
		{
			const LineDef * L = LineDefs[*it];

			if (L->Start()->x != L->End()->x) continue;
			if (L->Start()->y != L->End()->y) continue;

			// found a degenerate linedef
			count++;

fprintf(stderr, "LineDef[%d] is degenerate @ (%d,%d)\n",
*it, L->Start()->x, L->Start()->y);

			for (int which = 0; which < 2; which++)
			{
				int move_v  = (which == 0) ? L->start : L->end;
				int other_v = (which == 0) ? L->end : L->start;

				// don't move a vertex was not originally selected
				if (old_x.find(move_v) == old_x.end())
					continue;

				int nx = old_x[move_v];
				int ny = old_y[move_v];

				int dx = nx - Vertices[other_v]->x;
				int dy = ny - Vertices[other_v]->y;

				// this can only happen if a zero-length line was
				// part of the selection
				if (dx == 0 && dy == 0)
					dx = which ? -1 : 1;

				move_coord_for_quant(&nx, &ny, dx, dy, pass);

				BA_ChangeVT(*it, Vertex::F_X, grid.ForceSnapX(nx));
				BA_ChangeVT(*it, Vertex::F_Y, grid.ForceSnapY(ny));
			}
		}

		if (count == 0)
			break;
	}
#endif
}


void CMD_QuantizeObjects()
{
	selection_c list;
	selection_iterator_c it;

	if (! GetCurrentObjects(&list))
	{
		Beep();
		return;
	}

	BA_Begin();

	switch (edit.obj_type)
	{
		case OBJ_THINGS:
		{
			for (list.begin(&it); !it.at_end(); ++it)
			{
				const Thing * T = Things[*it];

				BA_ChangeTH(*it, Thing::F_X, grid.ForceSnapX(T->x));
				BA_ChangeTH(*it, Thing::F_Y, grid.ForceSnapY(T->y));
			}
			break;
		}

		case OBJ_RADTRIGS:
		{
			for (list.begin(&it); !it.at_end(); ++it)
			{
				const RadTrig * R = RadTrigs[*it];

				BA_ChangeRAD(*it, RadTrig::F_MX, grid.ForceSnapX(R->mx));
				BA_ChangeRAD(*it, RadTrig::F_MY, grid.ForceSnapY(R->my));

				int new_rw = grid.step * ((R->rw + grid.step - 1) / grid.step);
				int new_rh = grid.step * ((R->rh + grid.step - 1) / grid.step);

				BA_ChangeRAD(*it, RadTrig::F_RW, new_rw);
				BA_ChangeRAD(*it, RadTrig::F_RH, new_rh);
			}
			break;
		}

		case OBJ_VERTICES:
			DoQuantizeVertices(&list);
			break;

		// everything else merely quantizes vertices
		default:
		{
			selection_c verts(OBJ_VERTICES);
			ConvertSelection(&list, &verts);

			DoQuantizeVertices(&verts);
			break;
		}
	}

	BA_End();
}


/*
   move (x, y) to a new position: rotate and scale around (0, 0)
   */

#if 0
inline void RotateAndScaleCoords (int *x, int *y, double angle, double scale)
{
	double r, theta;

	r = hypot ((double) *x, (double) *y);
	theta = atan2 ((double) *y, (double) *x);
	*x = (int) (r * scale * cos (theta + angle) + 0.5);
	*y = (int) (r * scale * sin (theta + angle) + 0.5);
}
#endif


/*
   rotate and scale a group of objects around the centre of gravity
   */

void RotateAndScaleObjects (int objtype, SelPtr obj, double angle, double scale)
{
#if 0  // FIXME RotateAndScaleObjects
	int    dx, dy;
	int    centerx, centery;
	SelPtr cur, vertices;

	if (obj == NULL)
		return;


	switch (objtype)
	{
		case OBJ_THINGS:
			centre_of_things (obj, &centerx, &centery);
			for (cur = obj; cur; cur = cur->next)
			{
				dx = Things[cur->objnum].x - centerx;
				dy = Things[cur->objnum].y - centery;
				RotateAndScaleCoords (&dx, &dy, angle, scale);
				Things[cur->objnum].x = centerx + dx;
				Things[cur->objnum].y = centery + dy;
			}
			MarkChanges();
			break;

		case OBJ_VERTICES:
			centre_of_vertices (obj, &centerx, &centery);
			for (cur = obj; cur; cur = cur->next)
			{
				dx = Vertices[cur->objnum].x - centerx;
				dy = Vertices[cur->objnum].y - centery;
				RotateAndScaleCoords (&dx, &dy, angle, scale);
				Vertices[cur->objnum].x = (centerx + dx + /*4*/ 2) & ~/*7*/3;
				Vertices[cur->objnum].y = (centery + dy + /*4*/ 2) & ~/*7*/3;
			}
			MarkChanges(2);
			break;

		case OBJ_LINEDEFS:
			vertices = list_vertices_of_linedefs (obj);
			RotateAndScaleObjects (OBJ_VERTICES, vertices, angle, scale);
			ForgetSelection (&vertices);
			break;

		case OBJ_SECTORS:

			vertices = list_vertices_of_sectors (obj);
			RotateAndScaleObjects (OBJ_VERTICES, vertices, angle, scale);
			ForgetSelection (&vertices);
			break;
	}
#endif
}




/*
 *  exchange_objects_numbers
 *  Exchange the numbers of two objects
 *
 *  Return 0 on success, non-zero on failure.
 */
int exchange_objects_numbers (int obj_type, SelPtr list, bool adjust)
{
#if 0  // FIXME exchange_objects_numbers
	int n1, n2;

	// Must have exactly two objects in the selection
	if (list == 0 || list->next == 0 || (list->next)->next != 0)
	{
		BugError("exchange_object_numbers: wrong objects count.");
		return 1;
	}
	n1 = list->objnum;
	n2 = (list->next)->objnum;

	if (obj_type == OBJ_LINEDEFS)
	{
		struct LineDef swap_buf;
		swap_buf = LineDefs[n1];
		LineDefs[n1] = LineDefs[n2];
		LineDefs[n2] = swap_buf;
	}
	else if (obj_type == OBJ_SECTORS)
	{
		struct Sector swap_buf;
		swap_buf = Sectors[n1];
		Sectors[n1] = Sectors[n2];
		Sectors[n2] = swap_buf;
		if (adjust)
		{
			for (int n = 0; n < NumSideDefs; n++)
			{
				if (SideDefs[n].sector == n1)
					SideDefs[n].sector = n2;
				else if (SideDefs[n].sector == n2)
					SideDefs[n].sector = n1;
			}
		}
	}
	else if (obj_type == OBJ_THINGS)
	{
		struct Thing swap_buf;
		swap_buf = Things[n1];
		Things[n1] = Things[n2];
		Things[n2] = swap_buf;
	}
	else if (obj_type == OBJ_VERTICES)
	{
		struct Vertex swap_buf;
		swap_buf = Vertices[n1];
		Vertices[n1] = Vertices[n2];
		Vertices[n2] = swap_buf;
		if (adjust)
		{
			for (int n = 0; n < NumLineDefs; n++)
			{
				if (LineDefs[n].start == n1)
					LineDefs[n].start = n2;
				else if (LineDefs[n].start == n2)
					LineDefs[n].start = n1;
				if (LineDefs[n].end == n1)
					LineDefs[n].end = n2;
				else if (LineDefs[n].end == n2)
					LineDefs[n].end = n1;
			}
		}
	}
#endif
	return 0;
}


//--- editor settings ---
// vi:ts=4:sw=4:noexpandtab
